﻿module Fp10 where
import Data.Maybe (fromMaybe)
import Control.Monad.Writer
import Control.Monad.Reader
import Control.Monad.State

-----------------------------------------------------
-- Монада IO

main = do
  putStrLn "What is your name?"
  name <- getLine'
  putStrLn ("Nice to meet you, " ++ name ++ "!")

main' =
  putStrLn "What is your name?" >>
  getLine' >>= \name ->
  putStrLn $ "Nice to meet you, " ++ name ++ "!"


getLine' :: IO String
getLine' = do 
  c <- getChar
  if c == '\n' then
    return []
  else do 
    cs <- getLine'
    return (c:cs)
    
putStr'         :: String -> IO ()
putStr' []      =  return ()
putStr' (x:xs)  =  putChar x >> putStr' xs
    
putStr'' :: String -> IO ()
putStr'' = sequence_ . map putChar

putStr''' :: String -> IO ()
putStr''' = mapM_ putChar

-----------------------------------------------------
-- Монада Reader

-- простейший Reader (instance Monad ((->) e) where ...)
doIt :: Int -> Int
doIt = do
  a <- (^2)
  b <- (*5)
  return $ a + b
{-
*Fp10> doIt 3
24
-}

-- тип окружения - r
simpleReader :: Show r => Reader r String
simpleReader = reader (\e -> "Environment is " ++ show e)

-- *Fp10> runReader simpleReader 42
-- "Environment is 42"

-- *Fp10> runReader ask "Hello!"
-- "Hello!"

type User = String
type Password = String
type UsersTable = [(User,Password)]

pwds :: UsersTable
pwds = [("Bill","123"),("Ann","qwerty"),("John","2sRq8P")]

-- возвращает имя первого пользователя в списке
firstUser :: Reader UsersTable User
firstUser = do
  e <- ask
  let name = fst (head e)
  return name

-- *Fp10> runReader firstUser pwds
-- "Bill"
-- *Fp10> runReader firstUser []
-- "*** Exception: Prelude.head: empty list

-- возвращает длину пароля пользователя  или -1, если такого пользователя нет
getPwdLen :: User -> Reader UsersTable Int
getPwdLen person = do
  mbPwd <- asks $ lookup person 
  let mbLen = fmap length mbPwd
  let len = fromMaybe (-1) mbLen
  return len

-- *Fp10> runReader (getPwdLen "Ann") pwds
-- 6
-- *Fp10> runReader (getPwdLen "Ann") []
-- -1


usersCount :: Reader UsersTable Int
usersCount = asks length

localTest :: Reader UsersTable (Int,Int)
localTest = do
  count1 <- usersCount 
  count2 <- local (("Mike","1"):) usersCount 
  return (count1, count2)

-- *Fp10> runReader localTest pwds
-- (3,4)


-----------------------------------------------------
-- Монада Writer

-- runWriter (return 3 :: Writer String Int)  

type Vegetable = String
type Price = Double
type Qty = Double
type Cost = Double
type PriceList = [(Vegetable,Price)]

prices :: PriceList
prices = [("Potato",13),("Tomato",55),("Apple",48)]

-- tell позволяет задать вывод
addVegetable :: Vegetable -> Qty -> Writer (Sum Cost) (Vegetable, Price)
addVegetable veg qty = do
  let pr = fromMaybe 0 $ lookup veg prices
  let cost = qty * pr
  tell $ Sum cost
  return (veg, pr)
-- *Fp10> runWriter $ addVegetable "Apple" 100
-- (("Apple",48.0),Sum {getSum = 4800.0})
  
-- суммарная стоимость копится <<за кадром>>
myCart0 = do
  x1 <- addVegetable "Potato" 3.5
  x2 <- addVegetable "Tomato" 1.0
  x3 <- addVegetable "AGRH!!" 1.6
  return [x1,x2,x3]
-- *Fp10> runWriter myCart0
-- ([("Potato",13.0),("Tomato",55.0),("AGRH!!",0.0)],Sum {getSum = 100.5})

-- если хотим знать промежуточные стоимости, используем listen ...
myCart1 = do
  x1 <- listen $ addVegetable "Potato" 3.5
  x2 <- listen $ addVegetable "Tomato" 1.0
  x3 <- listen $ addVegetable "AGRH!!" 1.6
  return [x1,x2,x3]
-- *Fp10> runWriter myCart1
-- ([(("Potato",13.0),Sum {getSum = 45.5}),(("Tomato",55.0),Sum {getSum = 55.0}),((
-- "AGRH!!",0.0),Sum {getSum = 0.0})],Sum {getSum = 100.5})

-- ... или listens
myCart1' = do
  x1 <- listens getSum $ addVegetable "Potato" 3.5
  x2 <- listens getSum $ addVegetable "Tomato" 1.0
  x3 <- listens getSum $ addVegetable "AGRH!!" 1.6
  return [x1,x2,x3]
-- *Fp10> runWriter myCart1'
-- ([(("Potato",13.0),45.5),(("Tomato",55.0),55.0),(("AGRH!!",0.0),0.0)],Sum {getSum = 100.5})

--  (pass технический хелпер для censor)
--  для модификации лога используют censor :: (w -> w) -> Writer a -> Writer a
myCart0' = censor (discount 10) myCart0

-- бизнес-логика:)
discount proc (Sum x) = Sum $ if x < 100 then x
                              else x * (100 - proc) / 100
-- *Fp10> runWriter myCart0'
-- ([("Potato",13.0),("Tomato",55.0),("AGRH!!",0.0)],Sum {getSum = 90.45})

-----------------------------------------------------
-- Монада State

-- runState (return 3 :: State String Int) "Hi from State!"

tick :: State Int Int
tick = do 
  n <- get
  put (n+1)
  return n

succ' :: Int -> Int
succ' n = execState tick n

plus :: Int -> Int -> Int
plus n x = execState (sequence $ replicate n tick) x

plus' :: Int -> Int -> Int
plus' n x = execState (replicateM n tick) x
 
 